/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

#ifndef _COZ_AGENT_H
#define _COZ_AGENT_H

extern "C" void* consoleReaderThread(void* pCoZAgent);
extern "C" void* orphanDetectionThread(void* unused);
extern "C" void sig_alarm_handler(int sig_type);

/**
 * When a CoZLauncher is run as a job step on z/OS, it starts a CoZAgent on the
 * target host.  This program provides a command and control mechanism for the
 * application specific target program that will run on the target host.  The
 * CoZAgent is responsible for:
 *  - Establishing the client environment (via environment variables sent from 
 *    the job step) 
 *  - Starting the target program
 *  - Establishing data sources and sinks to tie the target program's 
 *    stdin, stdout and stderr to the job step's DD:SYSIN, DD:STDOUT and DD:STDERR 
 *    repectively
 *  - Handling MVS console MODIFY/STOP commands to control the target program.
 *    (What exactly are we going to do with MODIFY commands?)
 */
class CoZAgent {
	
private:
	//Thread functions
	friend void* consoleReaderThread(void* pCoZAgent);

	bool 		read_console;
	bool		debug_active;
	bool		debug_enabled;
	bool		quiet_mode;
	bool		no_shell;     // run target_command without a shell
	char* 		fullpath;
	char*		progname;
	char* 		target_command;
	int			child_wait_seconds;

	pid_t 		target_pid;
	pid_t 		stdin_dd_pid;
	pid_t 		stdout_dd_pid;
	pid_t 		stderr_dd_pid;
	
	pthread_t	console_tid;
	pthread_t	orphan_detection_tid;
	
	int*		stdin_dd_pipe;
	int*		stdout_dd_pipe;
	int*		stderr_dd_pipe;
		
public:
	CoZAgent(char* cmd_path);
	~CoZAgent();
	
	int run();
	void setReadConsole(bool b) { read_console = b; }
	void setDebugEnabled(bool b) { debug_enabled = b; }
	bool isStdinDrained() { return !isChildActive(stdin_dd_pid); }
	void setQuietMode(bool b) { quiet_mode = b; }
	void setChildWaitSeconds(int s) { child_wait_seconds = s; }
	void setNoShell(bool b) { no_shell = b; }

	static const int KEEP_ALIVE_DELAY_SECONDS = 5*60;

private:
	void configurePath();
	void initialize();
	bool isChildActive(pid_t pid);
	void killChild(const char* child_name, pid_t& pid, int signal);
	bool matchCommand(const char* cmd, char* buf);
	void setEnvironmentVariable(char* var);
	void startStdinDDReader();
	void startStdoutDDWriter();
	void startStderrDDWriter();
	void startTargetProgram();
	void usePipe(int* pipeFds, bool useReadEnd, bool useWriteEnd);
	int waitForChildren();
	int waitpidCleanup(const char* child_name, pid_t& pid, int status, bool ignoreSigPipe=false);
};

#endif  /*_COZ_AGENT_H*/
