/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

#ifndef _COZ_CLIENT_H
#define _COZ_CLIENT_H

#define _OPEN_SYS_SOCK_IPV6
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#include <netdb.h>
#include <sys/socket.h>

#include "CoZConstants.h"
#include "CoZPacket.h"


/**
 * A CoZClient object can connect to a dspipes-server over either ssh or socket
 * connection and through it execute a remote command, redirecting the command's
 * stdin/stdout/stderr output.
 *
 * The remote command is typically a fromdsn or todsn command, but
 * can actually be any command.
 *
 * The connection to the CoZServer is a full duplex binary connection,
 * which is either a tcp/ip socket or a ssh2 subsystem connection (which
 * are required to get clear binary data to/from z/OS sshd).
 *
 * Data sent to dspipes-server begins with a series of binary packets
 * which provide:
 *
 * - the clients id and codepage
 * - the remote command
 * - configure stdin/stdout/stderr channels for the remote command;
 *   either binary or text in the clients codepage
 *
 * the packets are followed by a stream of stdin data which is routed
 * by CoZServer to the remote command
 *
 * Data read from the CoZServer is a simple packet protocol
 * which multiplexes the stdout and stderr streams from the remote
 * command, as well as a special "error/status" packet which can
 * be used by some remote commands to relay encoded error/status information
 * to clients.
 *
 */

class CoZBaseClient {

protected:

	// Remote command arguments
    int 	cmd_argc;
    const char** 	cmd_argv;

    // Return code and server error/status info
    int		return_code;
    bool	server_exitcode_received;
    bool	server_exitstatus_received;
    int     server_retcode;
    int     server_errno;
    int     server_errno2;
    char*   server_msg;
    char*   server_errno_msg;
    char*   server_err_component;
    uint64_t total_bytes_sent;
    uint64_t total_bytes_received;
    uint64_t server_bytes_sent;
    uint64_t server_bytes_received;

	// switch to control ssh vs socket connection
	// The value of this switch is set to CONN_TYP_SSH
	// iff setSshOptions() is called
    static const int CONN_TYPE_SOCKET 	= 0;
    static const int CONN_TYPE_SSH 		= 1;
    int conn_type;

    static const size_t MAX_PADDR_LEN = 256;  // for getPrintedSocketAddr()
    static const int MAX_SOCKET_PATH_NAME = 2048;

    // IO Configuration options: see IO_CONFIG_* constants
	unsigned char io_config_stdin;
	unsigned char io_config_stdout;
	unsigned char io_config_stderr;
	unsigned char io_config_errpt;

	// ssh connection data
	char** ssh_cmd;   		// tokenized ssh command; built by setSshOptions()
	char*  user_host;
	bool   tunneled;

	// socket connection data
	int		sock;
	char*	socket_server_host;
	char* 	socket_server_port;
	char* 	socket_auth_token;

	// common stuff
    int     fd_conn_snd;  // to send control packets then stdin data to co-zServer
    int     fd_conn_rcv;  // to receive muxed command data from co-zServer

    char*	send_buf;
    char*	rcv_buf;

    char*   clientCodePage;

    int		serverProtocolVersion;
    int		protocolVersion;  // the negotiated protocol version

	char*   socket_file_name;

	static const int DEFAULT_CONNECT_TIMEOUT_SEC = 5;

	static const char* SERVER_SUBSYS_NAME;   // the ssh subsystem name of co-zServer


public:
	CoZBaseClient ();
	virtual ~CoZBaseClient();

	virtual int exec(const char* cmd_name, char** argv, int offset, int nargs) = 0;
	virtual void shutdownSocket() = 0;

	/** Configure IO channels; see IO_CONFIG_* constants */
	void setIOConfigStdin (unsigned const char v) { io_config_stdin  = v; };
	void setIOConfigStdout(unsigned const char v) { io_config_stdout = v; };
	void setIOConfigStderr(unsigned const char v) { io_config_stderr = v; };
	void setIOConfigErrpt (unsigned const char v) { io_config_errpt  = v; };

	char * setUserHost(char* _user_host);

	/** Return true if the client is using an ssh connection, false if using a socket */
	bool isSsh() { return conn_type == CONN_TYPE_SSH; }

	/** Set the host (or dotted addr name) if using socket; defaults to COZ_SERVER_HOST */
	void setSocketServerHost(const char* host) { socket_server_host = newStringCopy(host); }

	/** Set the "port" (or service name) if using socket; defaults to COZ_SERVER_PORT */
	void setSocketServerPort(const char* port) { socket_server_port = newStringCopy(port); }

	char * selectSocketFileName(bool quiet);

protected:
	void checkByteCounts();
	bool connectWithTimeout(struct sockaddr* address, socklen_t address_len);
    bool demuxPacket(int fd, char* packetBuf);
    void getPrintedSocketAddr(struct sockaddr* address, char* paddr);
	void negotiateProtocolVersion();
	void sendControlPackets();
	void sendClientInfoPacket();
	void sendIOConfigPacket();
	void sendAuthTokenPacket();
	void sendCommandPacket();
	void sendStreamBeginsPacket();
	void startSshConnection();
	unsigned short toNetUShort(unsigned short val);
	int32_t toNetInt(int32_t val);
	uint64_t toNetU64(uint64_t val);
	char* newStringCopy(const char*);
	int waitChild();
	void throwSigPipe();
	void reserveStandardHandles();
	int  sendCmdToLocalUnixDomainSocket(unsigned char);
	void setSocketFileName();


private:

	// prevent copy or assignment
	CoZBaseClient(CoZBaseClient const&);
	CoZBaseClient& operator=(CoZBaseClient const&);

	virtual void handleCmdIO() = 0;
	virtual void handleReceivedCommandPacket(CoZPacket* packet) = 0;

};


class CoZClient : public CoZBaseClient {

private:

    bool    user_set_stderr_option;

public:
	/** Construct an instance with the given remote command to execute */
	CoZClient();

	/** Destructor */
	~CoZClient();

	/** Execute the remote command via CoZServer */
	int exec(const char* cmd_name, char** argv, int offset, int nargs);
	void shutdownSocket();
	/** Set the connection type to SSH, and gives the args to the ssh command.
	 *  The last arg must be 'user@host' */
	void setSshOptions(char** argv, int istart, int nargs);
	void setUserSetStderrOption(bool b);

private:

    void handleCmdIO();
	void handleReceivedCommandPacket(CoZPacket* packet);
	void getSocketEnvironment();
	void setControlPathOptions();
	void startSocketConnection();

};

class CoZControlClient : public CoZBaseClient {

private:

	static const char* COZ_CONTROL_PATH;     // for client initiated tunneling case, default control path
	static const int DEFAULT_COZ_SSH_CONNECT_TIMEOUT_SEC = 30;  //Amount of time for cozcontrol to wait for an ssh connection
	static const int MAX_EPHEMERAL_PORT_ATTEMPTS = 3; // number of times to try local SSH forwarding using an ephemeral port

    int     local_sock;		// unix domain socket used for cozcontrol packets
    pid_t	foreground_pid; // pid of process prior to backgrounding
    char*   env_coz_control_path;
    char*   env_coz_server_host;
	char*   env_coz_server_port;
	char*   env_coz_auth_token;

public:
	CoZControlClient();

	/** Destructor */
	~CoZControlClient();

	/** Execute the remote command via CoZServer */
	int exec(const char* cmd_name, char** argv, int offset, int nargs);
	void shutdownSocket();
	/** Set the connection type to SSH, and gives the args to the ssh command.
	 *  The last arg must be 'user@host' */
	void setSshOptions(char** argv, int istart, int nargs, bool tunneling);

private:
    void handleCmdIO();
	void handleReceivedCommandPacket(CoZPacket* packet);
	int  startUnixDomainSocketListener();
	void runUnixDomainSocketServer(int listen_sock);
	void forkToBackground();
	uint16_t  getEphemeralPort();

};

#endif  /*_COZ_CLIENT_H*/
