/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */

//The following includes must be in this order for CoZ
#include "CoZConstants.h"
#include "DtlLog.h"
#include "sys_util.h"
#include "mfocus.h"

static const char ASCII_LF = (char)0x0A;
static const char ASCII_NEL = (char)0x85;
static const char EBC_NL = (char)0x15;
static const char EBC_LF = (char)0x25;
static const char EBC_CR = (char)0x0D;
static const char EBC_SP = (char)0x40;

#define Str(x) #x
#define Xstr(x) Str(x)
const char* COZ_VERSION = Xstr(BUILD_VERSION) ;  

/**
 * genlines - generate lines of data for testing
 * 
 */

void usage() {
		
	fprintf(stderr,
"usage: genlines [-l nl (default) | cr | lf | crlf | crnl | rdw | ibmrdw | mfrdw | 0xbb[bb] | none] minsize maxsize count \n"
"   or: genlines -h  -> display this help and exit\n" 
"   or: genlines -v  -> display the current version and exit\n"
	);
}

int main(int argc, char** argv) {
	
	char* pgm;
	DtlLog::setProgName(basename(argv[0]));

	unsigned long count;
	unsigned long minsize;
	unsigned long maxsize;
	    
    // now parse the switches	
	bool use_rdw = false;
	bool use_mf  = false;
	char lf = '\n';
	char nl = '\n';
	if (nl == EBC_NL) { //EBCDIC
		lf = EBC_NL;
		nl = EBC_LF;
	}
	char terminator[3] = {lf, '\0', '\0'};  // default is lf
		
	extern char* optarg;
	extern int optind;
	int opt;
	
	while ((opt = getopt(argc, argv, ":hl:v")) != -1) {
		switch (opt) {    
			case 'h':
    			usage();
    			exit(RC_OK);
    			break;
			case 'l':
				use_rdw = false;
				use_mf = false;
				if (!strcmp("cr", optarg)) {
					terminator[0] = '\r';
					terminator[1] = '\0';
				} else if (!strcmp("lf", optarg)) {
					terminator[0] = lf;
					terminator[1] = '\0';
				} else if (!strcmp("crlf", optarg)) {
					terminator[0] = '\r';
					terminator[1] = lf;
					terminator[2] = '\0';
				} else if (!strcmp("crnl", optarg)) {
					terminator[0] = '\r';
					terminator[1] = nl;
					terminator[2] = '\0';
				} else if (!strcmp("rdw", optarg) || !strcmp("ibmrdw", optarg)) {
					// rdw is now an alias for ibmrdw
					use_rdw = true;
					terminator[0] = '\0';
				} else if (!strcmp("mfrdw", optarg)) {
					use_mf = true;
					use_rdw = true;
					terminator[0] = '\0';
				} else if (!strncasecmp("0x", optarg, 2)) {
					const char* p = optarg+2;
					if (strlen(p) % 2) {
						LOG(LOG_WARNING, "Line termination hex string invalid: \"%s\"", optarg);
						return RC_INPUT_ERROR;
					}
					int nchars = strlen(p) / 2;
					if (nchars > 2) {
						LOG(LOG_WARNING, "Line termination hex string too long: \"%s\"", optarg);
						return RC_INPUT_ERROR;
					}
					unsigned long val;
					char* q;
					val = strtoul(p, &q, 16);
					if (q != p+strlen(p)) {
						LOG(LOG_ERR, "Line termination hex string invalid: \"%s\"", optarg);
		    			return RC_INPUT_ERROR;
					}
					memcpy(terminator, ((char*)&val)+sizeof(val)-nchars, nchars);
				} else if (!strcmp("none", optarg)) {
					terminator[0] = '\0';
				}
				break;
			case 'v':
    			printf("%s\n", COZ_VERSION);
    			exit(RC_OK);
    			break;
    		case ':':
				LOG(LOG_ERR, "Missing argument for option: %c", optopt);
				usage();
    			return RC_INPUT_ERROR;
				break;
			default:
				LOG(LOG_ERR, "Unknown option: %c", optopt);
				usage();
    			return RC_INPUT_ERROR;
    	}
    }
    
    // Got options.  The last three arguments are min/max/count;
    if (argc-optind != 3) {
    	usage();
    	return RC_INPUT_ERROR;
    } else {
    	minsize = atoi(argv[optind++]);
    	maxsize = atoi(argv[optind++]);
    	count = atoi(argv[optind]);
    }	
    
    if (minsize > maxsize) {
    	usage();
    	return RC_INPUT_ERROR;
    }
            
    char* buf = new char[maxsize+4];
    char digits[11] = "0123456789";
    
	unsigned short int seed[3];   // our original seed for lineLength
	unsigned short int randx[3];  // current/previous seed/generator value
    
	time_t tt = time(NULL);
	memcpy((void*)seed, (void*)&tt, sizeof(seed)); 
	seed[2] = 0x330E;  // same as srand48()
	memcpy((void*)randx, (void*)seed, sizeof(randx));
    
    int tlen = strlen(terminator);
    uint32_t* pRdw = (uint32_t*)buf; 
    
    uint32_t totalBytes = 0;
    uint32_t range = maxsize - minsize + 1;
    
    LOG(LOG_DEBUG, "termlen=%d, use_rdw=%d, use_mf=%d, minsize=%lu, maxsize=%lu, count=%d",
    		tlen, use_rdw, use_mf, minsize, maxsize, count);
    
    // write the micro-focus header
    if (use_mf) {
    	MF_HEADER mf_header;
    	init_mf_header(&mf_header, minsize, maxsize);
		fwrite((void*)&mf_header, sizeof(char), sizeof(mf_header), stdout);
    }
    
    char* write_ptr = buf;
    for (int i=0; i<count; i++) {
    	int len = minsize;
    	if (maxsize != minsize) {
			len = minsize + (nrand48(randx) % range);
    	}
		
		if (!use_rdw) {
			memset(buf, digits[i%10], len);
			if (tlen) {
				strcpy(buf+len, terminator);
				len += tlen;
			}
		} else if (use_mf) { // -l mfrdw
			if (maxsize <= MAX_MF_SHORT_REC_SIZE) {
				write_ptr = buf+2;
				uint16_t* pRdw2 = (uint16_t*)write_ptr;
				*pRdw2 = htons(len | 0x4000);
				memset(buf+4, digits[i%10], len);
				len += 2;
			} else {
				*pRdw = htonl(len | 0x40000000);
				memset(buf+4, digits[i%10], len);
				len += 4;			
			}
			// 4-byte align/pad data written with <= 3 bytes of zeros
			if (int excess = len % 4) {
				memset(write_ptr+len, 0, 4-excess);
				len += 4-excess;
			}
		} else { // -l rdw or ibmrdw
		    *pRdw = htonl((len+4)<<16);  // len includes RDW len
			memset(buf+4, digits[i%10], len);
			len += 4;			
		}
		fwrite(write_ptr, sizeof(char), len, stdout);
		totalBytes += len;
	}
	fflush(stdout);
	fclose(stdout);	
    LOG(LOG_NOTICE, "%lu lines / %lu bytes written to stdout", count, totalBytes);
    
    return RC_OK;
} 



