/*
 * Copyright 2009-2015 Dovetailed Technologies, LLC.
 *
 * Licensed under the accompanying Co:Z Target System Toolkit Source License Agreement (the "License");
 * you may not use this file except in compliance with the License.
 *
 */


//The following includes must be in this order for CoZ
#include "CoZConstants.h"
#include "DtlLog.h"
#include "sys_util.h"

#define Str(x) #x
#define Xstr(x) Str(x)
const char* COZ_VERSION = Xstr(BUILD_VERSION) ;

/**
 * toasa - convert ASCII form feeds to ASA control characters
 * 
 * Currently, this only converts ASCII form-feeds.  Output lines
 * will only have '1' (page eject) or ' ' (single line) carriage
 * control in column one of each output line.
 *
 */

void usage() {
		
	fprintf(stderr,
"usage: toasa     -> convert stdin to ASA compatible format.\n"
"   or: toasa -h  -> display this help and exit\n"
"   or: toasa -v  -> display the current version and exit\n"
"\n"
"NOTE: \n"
"  -Currently only converts ASCII form-feeds.  Output lines will\n"
"   have ASA character '1' (page eject) or ' ' (single line)\n"
"   carriage control in column one of each output line.\n"
"  -ASCII form-feed doesn't preserve the column position.\n"
	);

}

#ifndef HAVE_GETLINE
/**
 * a (safe) line reader, compatible with the GNU / POSIX_C_SOURCE=2008 implementation
 */
static ssize_t getline (char**ppBuf, size_t *pBufLen, FILE* stream) {
	TRACE_CTX_FINE("getline");
	static const size_t BUF_SIZE = 4096;

	if (*ppBuf == NULL) {
		*ppBuf = (char*)malloc(BUF_SIZE);
		if (*ppBuf == NULL) {
			perror("malloc() error");
			return -1;
		}
		*pBufLen = BUF_SIZE;
	}

	ssize_t rlen = -1;
	if (fgets(*ppBuf, *pBufLen, stream)) {
		rlen = strlen(*ppBuf);
		TRACEV("fgets() = %zd", rlen);
		while (rlen && (*ppBuf)[rlen-1] != '\n' && !feof(stream)) {
			// buffer was too small for entire line
			char* newBuf = (char*)realloc(*ppBuf, *pBufLen+BUF_SIZE);
			if (!newBuf) {
				perror("realloc() error");
				return -1;
			}
			*ppBuf = newBuf;
			*pBufLen += BUF_SIZE;
			fgets(*ppBuf+rlen, *pBufLen-rlen, stream);
			rlen = strlen(*ppBuf);
			TRACEV("fgets() = %zd", rlen);
		}
	}
	TRACE_CTX_RETVAL("%zd", rlen);
	return rlen;
}
#endif  /* HAVE_GETLINE */

/*
 * Write a line to stdout, followed by a \n
 * If the line starts with a form-feed, then
 * the first output column will contain '1',
 * otherwise the first output column will be ' '.
 *
 * Note: the input line may or may not be terminated by a newline character.
 */
static void writeLine(char* line, int len) {
	TRACE_CTX_ARGS_FINE("writeLine", "len=%d: %.*s", len, len, line);
	// determine the asa
	char asa;
	if (len > 0 && *line == '\f') {
		asa = '1';
		line++;
		len--;
	} else {
		asa = ' ';
	}
	// drop any trailing newline, since we will always add below
	if (len > 0 && line[len-1] == '\n') {
		len--;
	}

	if (fprintf(stdout, "%c%.*s\n", asa, len, line) < 0) {
		perror("error writing output");
	}
}

/*
 * Convert a line that may contain one or more form-feed characters
 * and write one or more lines containing ASA control characters to stdout.
 */
static void processLine(char* line, int len) {
	TRACE_CTX_ARGS_FINE("processLine", "len=%d: %.*s", len, len, line);
	char* pFF = NULL;
	while (len && (pFF = (char*)memchr(line, '\f', len)) != NULL) {
		// found a FF somewhere ...
		ssize_t lpiece;
		if (pFF > line) {
			// handle a non-empty piece prior to the FF
			lpiece = pFF - line;
			writeLine(line, lpiece);
			line += lpiece;
			len -= lpiece;
		}
		// now the line starts with FF
		// find the next piece up to another possible FF
		if (len && (pFF = (char*)memchr(line+1, '\f', len-1)) != NULL) {
			lpiece = pFF - line;
		} else {
			lpiece = len;
		}
		writeLine(line, lpiece);
		line += lpiece;
		len -= lpiece;
	}
	// a piece that doesn't have a FF in it
	if (len) {
		writeLine(line, len);
	}
}

int main(int argc, char** argv) {
	size_t lbuf_len;
	char* lbuf = NULL;
	ssize_t line_len = 0;
	
	DtlLog::setProgName(basename(argv[0]));

	extern char* optarg;
	extern int optind;
	int opt;

	while ((opt = getopt(argc, argv, ":hv")) != -1) {
		switch (opt) {
			case 'h':
    			usage();
    			exit(RC_OK);
    			break;
			case 'v':
    			printf("%s\n", COZ_VERSION);
    			exit(RC_OK);
    			break;
			default:
				LOG(LOG_ERR, "Unknown option: %c", optopt);
				usage();
    			return RC_INPUT_ERROR;
    	}
    }

	while ((line_len = getline(&lbuf, &lbuf_len, stdin)) >= 0) {
		processLine(lbuf, (size_t)line_len);
	}
	if (ferror(stdin)) {
		perror("error reading input");
		return 2;
	} else {
		return 0;
	}
} 

